# Grab the type definitions from /usr/include/bits/utmp.h

foreach {name num} {
         EMPTY           0     RUN_LVL         1
         BOOT_TIME       2     NEW_TIME        3
         OLD_TIME        4     INIT_PROCESS    5
         LOGIN_PROCESS   6     USER_PROCESS    7
         DEAD_PROCESS    8     ACCOUNTING      9 } {
    set types($num) $name
}

# We'll use $zero to trim trailing zeros from the data.
set zero [binary format c 0x00]

# $reportFmt defines the widths of the columns in the report
set reportFmt {%-16s %6s %12s %20s %-12s %6s %5s %7s %21s %16s}

set output stdout

set mtime 0
while {1} {
    while {[file mtime /var/run/utmp] == $mtime} {
        after 10000
    }
    set mtime [file mtime /var/run/utmp]

    # Display a header
    puts [format $reportFmt Type Pid Line ID User Host Exit Session Time Addr]

    # Open read and close the utmp file
    set if [open /var/run/utmp r]
    set d [read $if]
    close $if

    # Save the length of the data buffer for future use.
    set utmpLen [string length $d]

    # This is the start of the utmp structure being examined
    set start 0

    # As long as there is data to parse, parse it and step to the
    #   next structure.

    while {$start < $utmpLen} {
        # start of struct  type padding          pid line id user host exit\
          session time addr
        set fmt "@$start    s  @[expr $start +4] i   a32  a4 a32  a256 s2  \
          i       i2   c4"
        binary scan $d $fmt type pid line id user host exit session time addr

        # Trim trailing zeros
        foreach v {line id user host} {
            set $v [string trim [set $v] $zero]
        }

        puts $output [format $reportFmt \
          $types($type) $pid $line $id $user $host $exit $session \
          [clock format [lindex $time 0] -format "%D %r"] [join $addr "."]]

        incr start 384
    }
}

